@@ -11,7 +11,7 @@ from enum import Enum
 from typing import NamedTuple
 from uuid import UUID
 
-from gi.repository import GLib, GObject
+from gi.repository import GLib, GObject, Gio, Gtk
 from pyotp import OTP, TOTP, parse_uri
 
 if typing.TYPE_CHECKING:
@@ -33,7 +33,7 @@ class EntryColor(Enum):
     BROWN = "BrownColorButton"
 
 
-class SafeElement(GObject.GObject):
+class SafeElement(GObject.Object):
 
     sorted_handler_id: int | None = None
 
@@ -62,12 +62,13 @@ class SafeElement(GObject.GObject):
 
         return False
 
-    @GObject.Signal()
+    @GObject.Signal(flags=GObject.SignalFlags.ACTION)
     def updated(self):
         """Signal used to tell whenever there have been any changed that should
         be reflected on the main list box or edit page."""
         self._db_manager.is_dirty = True
         self.touch(modify=True)
+        self.emit(self.updated)
         logging.debug("Safe element updated")
 
     def touch(self, modify: bool = False) -> None:
@@ -78,35 +79,79 @@ class SafeElement(GObject.GObject):
     def delete(self) -> None:
         """Delete an Element from the database."""
         element = self._element
-        uuid = self.uuid
         parentgroup = self.parentgroup
         if self.is_entry:
             self._db_manager.db.delete_entry(element)
+            found, pos = self._db_manager.entries.find(self)
+            if found:
+                self._db_manager.entries.remove(pos)
         else:
             self._db_manager.db.delete_group(element)
+            found, pos = self._db_manager.groups.find(self)
+            if found:
+                self._db_manager.groups.remove(pos)
 
         parentgroup.updated()
-        self._db_manager.emit("element-removed", uuid)
+
+    def trash(self) -> bool:
+        """Delete an Element from the database.
+
+        Returns if the element was deleted.
+        """
+        element = self._element
+        parentgroup = self.parentgroup
+        if parentgroup.is_trash_bin:
+            self.delete()
+            return True
+
+        if self.is_trash_bin:
+            self.delete()
+            self._db_manager.trash_bin = None
+            return True
+
+        trash_bin_missing = SafeGroup.get_trash_bin(self._db_manager) is None
+
+        if self.is_entry:
+            self._db_manager.db.trash_entry(element)
+            found, pos = self._db_manager.entries.find(self)
+            if found:
+                self._db_manager.entries.items_changed(pos, 1, 1)
+        else:
+            self._db_manager.db.trash_group(element)
+            found, pos = self._db_manager.groups.find(self)
+            if found:
+                self._db_manager.groups.items_changed(pos, 1, 1)
+
+        # We add the trash bin if it was not present already
+        if trash_bin_missing:
+            trash_bin = SafeGroup.get_trash_bin(self._db_manager)
+            self._db_manager.groups.append(trash_bin)
+
+        parentgroup.updated()
+        return False
 
     def move_to(self, dest: SafeGroup) -> None:
-        old_location = self.parentgroup.uuid
-        new_location = dest.uuid
+        old_location = self.parentgroup
 
-        if self.parentgroup == dest:
+        if old_location == dest:
             return
 
-        # TODO: we will crash if uuid does not exist
+        # We signal items changed on the main list store so that the changes are
+        # propagated to the filter models used by each group.
         if self.is_entry:
             self._db_manager.db.move_entry(self._element, dest.group)
+            found, pos = self._db_manager.entries.find(self)
+            if found:
+                self._db_manager.entries.items_changed(pos, 1, 1)
         else:
             self._db_manager.db.move_group(self._element, dest.group)
+            found, pos = self._db_manager.groups.find(self)
+            if found:
+                self._db_manager.groups.items_changed(pos, 1, 1)
 
-        self.parentgroup.updated()
+        old_location.updated()
         dest.updated()
 
-        # We use the UUIDs stored before moving the element.
-        self._db_manager.emit("element-moved", self, old_location, new_location)
-
     @property
     def element(self) -> Entry | Group:
         return self._element
@@ -165,8 +210,28 @@ class SafeElement(GObject.GObject):
         if self.is_root_group:
             return self
 
+        for group in self._db_manager.groups:
+            if group.uuid == self._element.parentgroup.uuid:
+                return group
+
+        logging.error("This should be unreachable: parentgroup")
         return SafeGroup(self._db_manager, self._element.parentgroup)
 
+    @property
+    def parentgroup_uuid(self) -> UUID:
+        """UUID of the parent Group of the element
+
+        This method should be preferred than parentgroup since it does not go
+        through the entire list of elements.
+
+        :returns: parent group
+        :rtype: SafeGroup
+        """
+        if self.is_root_group:
+            return self.uuid
+
+        return self._element.parentgroup.uuid
+
     @property
     def is_root_group(self) -> bool:
         if self.is_entry:
@@ -174,6 +239,13 @@ class SafeElement(GObject.GObject):
 
         return self._element.is_root_group
 
+    @property
+    def is_trash_bin(self) -> bool:
+        if (trash_bin_inner := self._db_manager.db.recyclebin_group):
+            return self.uuid == trash_bin_inner.uuid
+
+        return False
+
     @property
     def uuid(self) -> UUID:
         """UUID of the element
@@ -225,6 +297,10 @@ class SafeElement(GObject.GObject):
 
 
 class SafeGroup(SafeElement):
+
+    _entries = None
+    _subgroups = None
+
     def __init__(self, db_manager: DatabaseManager, group: Group) -> None:
         """GObject to handle a safe group.
 
@@ -239,8 +315,25 @@ class SafeGroup(SafeElement):
     def get_root(db_manager: DatabaseManager) -> SafeGroup:
         """Method to obtain the root group."""
         # pylint: disable=no-member
+        for group in db_manager.groups:
+            if group.is_root_group:
+                return group
+
+        logging.error("This should be unreachable: get_root")
         return SafeGroup(db_manager, db_manager.db.root_group)
 
+    @staticmethod
+    def get_trash_bin(db_manager: DatabaseManager) -> SafeGroup | None:
+        if (trash_bin := db_manager.trash_bin):
+            return trash_bin
+
+        if (trash_bin_inner := db_manager.db.recyclebin_group):
+            trash_bin = SafeGroup(db_manager, trash_bin_inner)
+            db_manager.trash_bin = trash_bin
+            return trash_bin
+
+        return None
+
     def new_entry(
         self, title: str = "", username: str = "", password: str = ""
     ) -> SafeEntry:
@@ -262,7 +355,7 @@ class SafeGroup(SafeElement):
         )
         safe_entry = SafeEntry(self._db_manager, new_entry)
         self.updated()
-        self._db_manager.emit("element-added", safe_entry, self.uuid)
+        self._db_manager.entries.append(safe_entry)
 
         return safe_entry
 
@@ -275,32 +368,54 @@ class SafeGroup(SafeElement):
         )
         safe_group = SafeGroup(self._db_manager, new_group)
         self.updated()
-        self._db_manager.emit("element-added", safe_group, self.uuid)
+        self._db_manager.groups.append(safe_group)
 
         return safe_group
 
     @property
-    def subgroups(self) -> list[SafeGroup]:
-        return [SafeGroup(self._db_manager, group) for group in self._group.subgroups]
+    def subgroups(self) -> Gio.ListModel:
+        if self._subgroups is None:
+            filter_ = Gtk.CustomFilter.new(self._group_filter_func)
+            model = Gtk.FilterListModel.new(self._db_manager.groups, filter_)
+            self._subgroups = model
+
+        return self._subgroups
 
     @property
-    def entries(self) -> list[SafeEntry]:
-        return [SafeEntry(self._db_manager, entry) for entry in self._group.entries]
+    def entries(self) -> Gio.ListModel:
+        if self._entries is None:
+            filter_ = Gtk.CustomFilter.new(self._group_filter_func)
+            model = Gtk.FilterListModel.new(self._db_manager.entries, filter_)
+            self._entries = model
+
+        return self._entries
 
     @property
     def group(self) -> Group:
         """Returns the private pykeepass group."""
         return self._group
 
+    def _entry_filter_func(self, entry: SafeEntry) -> bool:
+        return entry.parentgroup_uuid == self.uuid
+
+    def _group_filter_func(self, group: SafeGroup) -> bool:
+        if group.is_root_group:
+            return False
+
+        return group.parentgroup_uuid == self.uuid
+
 
 class SafeEntry(SafeElement):
     # pylint: disable=too-many-instance-attributes, too-many-public-methods
 
     _color_key = "color_prop_LcljUMJZ9X"
+    _expired_id: int | None = None
     _note_key = "Notes"
     _otp: OTP | None = None
     _otp_key = "otp"
 
+    history_saved = GObject.Signal()
+
     def __init__(self, db_manager: DatabaseManager, entry: Entry) -> None:
         """GObject to handle a safe entry.
 
@@ -313,11 +428,20 @@ class SafeEntry(SafeElement):
 
         self._attachments: list[Attachment] = entry.attachments or []
 
-        self._attributes: dict[str, str] = {
-            key: value
-            for key, value in entry.custom_properties.items()
-            if key not in (self._color_key, self._note_key, self._otp_key)
-        }
+        # NOTE Can fail at libpykeepass, see
+        # https://github.com/libkeepass/pykeepass/issues/254
+        # TODO Read as many attributes as possible until we see an error.
+        try:
+            attributes = {
+                key: value
+                for key, value in entry.custom_properties.items()
+                if key not in (self._color_key, self._note_key, self._otp_key)
+            }
+        except Exception as err:  # pylint: disable=broad-except
+            logging.error("Could not read attributes: %s", err)
+            attributes = {}
+
+        self._attributes: dict[str, str] = attributes
 
         color_value: str = entry.get_custom_property(self._color_key)
         self._color: str = color_value or EntryColor.NONE.value
@@ -333,8 +457,7 @@ class SafeEntry(SafeElement):
             except ValueError as err:
                 logging.debug(err)
 
-        # Check if the entry has expired every 10 minutes.
-        GLib.timeout_add_seconds(600, self._is_expired)
+        self._check_expiration()
 
     @property
     def entry(self) -> Entry:
@@ -377,12 +500,48 @@ class SafeEntry(SafeElement):
         safe_entry = SafeEntry(self._db_manager, clone_entry)
 
         self.parentgroup.updated()
-        self._db_manager.emit("element-added", safe_entry, safe_entry.parentgroup.uuid)
+        self._db_manager.entries.append(safe_entry)
+
+    def _check_expiration(self) -> None:
+        """Check expiration
+
+        If the entry is expired, this ensures that a notification is sent.
+        If the entry is not expired yet, a timeout is set to regularly
+        check if the entry is expired.
+        """
+        if self._expired_id:
+            GLib.source_remove(self._expired_id)
+            self._expired_id = None
+
+        if not self.props.expires:
+            return
+
+        if self.props.expired:
+            self.notify("expired")
+        else:
+            self._expired_id = GLib.timeout_add_seconds(600, self._is_expired)
 
     def _is_expired(self) -> bool:
-        self.props.expired = self.element.expired
+        if self.props.expired:
+            self._expired_id = None
+            self.notify("expired")
+            return GLib.SOURCE_REMOVE
+
         return GLib.SOURCE_CONTINUE
 
+    def save_history(self) -> None:
+        """Save current version of the entry in its history."""
+        # NOTE Attachments are references, so duplicating them is ok.
+        self._entry.save_history()
+        self.updated()
+        self.emit(self.history_saved)
+
+    def delete_history(self, entry: SafeEntry) -> None:
+        """Delete entry from the history of self."""
+        self._entry.delete_history(entry.entry)
+        self.updated()
+        self.emit(self.history_saved)
+
     @GObject.Property(type=object, flags=GObject.ParamFlags.READABLE)
     def attachments(self) -> list[Attachment]:
         return self._attachments
@@ -399,6 +558,7 @@ class SafeEntry(SafeElement):
         attachment = self._entry.add_attachment(attachment_id, filename)
         self._attachments.append(attachment)
         self.updated()
+        self.notify("attachments")
 
         return attachment
 
@@ -409,6 +569,7 @@ class SafeEntry(SafeElement):
         """
         self._db_manager.db.delete_binary(attachment.id)
         self._attachments.remove(attachment)
+        self.notify("attachments")
         self.updated()
 
     def get_attachment(self, id_: str) -> Attachment | None:
@@ -448,9 +609,13 @@ class SafeEntry(SafeElement):
         :param str key: attribute key
         :param str value: attribute value
         """
+        if self.props.attributes.get(key) == value:
+            return
+
         self._entry.set_custom_property(key, value)
         self._attributes[key] = value
         self.updated()
+        self.notify("attributes")
 
     def delete_attribute(self, key: str) -> None:
         """Delete an attribute
@@ -463,6 +628,7 @@ class SafeEntry(SafeElement):
         self._entry.delete_custom_property(key)
         self._attributes.pop(key)
         self.updated()
+        self.notify("attributes")
 
     @GObject.Property(type=str, default=EntryColor.NONE.value)
     def color(self) -> str:
@@ -484,6 +650,11 @@ class SafeEntry(SafeElement):
             self._entry.set_custom_property(self._color_key, new_color)
             self.updated()
 
+    @property
+    def history(self) -> list[SafeEntry]:
+        history = self._entry.history
+        return [SafeEntry(self._db_manager, entry) for entry in history]
+
     @GObject.Property(type=object)
     def icon(self) -> Icon:
         """Get icon number
@@ -535,7 +706,8 @@ class SafeEntry(SafeElement):
         if not otp and self._otp:
             # Delete existing
             self._otp = None
-            self._element.otp = None
+            # NOTE the opt property doesn't accept None.
+            self._element.otp = ""
             self.updated()
         elif self._otp and self._otp.secret != otp:
             # Changing an existing OTP
@@ -645,10 +817,12 @@ class SafeEntry(SafeElement):
     def expires(self, value: bool) -> None:
         if value != self.entry.expires:
             self.entry.expires = value
-            self.props.expired = self.entry.expired
+            self._check_expiration()
             self.updated()
 
-    @GObject.Property(type=bool, default=False)
+    @GObject.Property(
+        type=bool, default=False, flags=GObject.ParamFlags.READABLE
+        | GObject.ParamFlags.EXPLICIT_NOTIFY)
     def expired(self):
         return self.entry.expired
 
@@ -681,7 +855,8 @@ class SafeEntry(SafeElement):
                 tzinfo=timezone.utc,
             )
             self.entry.expiry_time = expired
-            self.props.expired = self.entry.expired
+            self._check_expiration()
+
             self.updated()
 
 
